// JabberingThreads.java
//
// CompSci 141 / CSE 141 / Informatics 101 Spring 2013
// Code Example
//
// This is a completed version of the "jabbering threads" example from
// lecture.  The program takes as many command-line arguments as you want
// to give it (though it must be given at least two).  For example, you
// could run the program from the command line like this:
//
//     java JabberingThreads alex is happy today
//
// It is also possible, of course, to run this program via Eclipse, supplying
// command-line parameters to it in the usual way.
//
// For each word you pass on the command line, a thread is created that
// will print that word when it is allowed to speak.  Only one thread may
// speak at a time.  A combination of wait(), notify(), and a synchronized
// block are used to ensure that exactly one thread can speak at a time.
// To keep the program from overwhelming us with output, each thread sleeps
// for a quarter of a second after stating its word but before it allows the
// next thread to speak.
//
// The output of the program may vary.  I ran it on the Oracle JVM (1.7.0)
// on Windows 7 Professional (64-bit) and, interestingly, notify() woke threads
// in order, so it looks like the combination of this VM and Windows 7
// implements the wait/notify mechanism as a first-in-first-out "fair"
// one, such that the output of the program looked like this:
//
//     alex is happy today alex is happy today alex ...
//
// Running the program on different virtual machines and/or different operating
// systems may display the words in different orders, though it should always
// be true that exactly one thread will speak at a time.

import java.util.ArrayList;


public class JabberingThreads
{
	public static void main(String[] args)
	{
		if (args.length < 2)
		{
			// If there are fewer than two command-line arguments, print an
			// error message and exit the program.
	
			System.out.println("This program requires at least two command-line arguments.");
			return;

			// For fun, try removing this restriction and pass only one
			// command-line argument to the program, just to see what will
			// happen.  (For even more fun, first try to guess what will
			// happen, then see if you're right.  You might as well challenge
			// yourself!)
		}

		System.out.println("Starting jabberer threads...");


		// This is the shared object that all the threads use for wait/notify.
		// By having all the threads wait/notify on the same object, only one
		// thread will be allowed to speak at a time.  (This is explained
		// further in the comments in the Jabberer class.)
		//
		// The reason I've created an object of type Object is because all
		// objects in Java have monitors in them, which means that they all
		// support wait/notify.  Since the only thing I need this object for
		// is wait/notify, there's no point in using any other class: Object
		// will do just fine.

		Object monitor = new Object();


		// Now we'll create an array of threads, one for each command-line
		// argument.  Each thread will manage a Jabberer object.  Each will
		// be given a different command-line argument, but all will be
		// passed the same monitor.

		ArrayList<Thread> threads = new ArrayList<Thread>();
		
		for (String s : args)
		{
			threads.add(new Thread(new Jabberer(s, monitor)));
		}
		
		for (Thread t : threads)
		{
			t.start();
		}


		// This is cheesy, but it works.  To make sure all the threads have
		// had a chance to start before we start the jabberers, the program
		// will sleep for three seconds after calling the start() method on
		// all the threads.  After that length of time, all the jabberers
		// should be waiting on the monitor.  (While there's no guarantee
		// that this is true, I'd be truly shocked if any of you have a
		// processor slow enough that it can't get a few threads started
		// in three full seconds!)
		//
		// A better approach would be to employ some kind of communication
		// back to the main thread, but that seemed like overkill for this
		// relatively straightforward example.

		try
		{
			System.out.println("Waiting for 3 seconds to allow threads to start...");
			Thread.sleep(3000);
		}
		catch (InterruptedException e)
		{
		}

		System.out.println("Starting the jabberers jabbering for 15 seconds...");


		// Now we'll get the ball rolling by notifying the monitor.  This will
		// wake exactly one of the jabberers (all of whom are presently waiting
		// on the monitor), but we don't know which one.
		//
		// Now's a good time to briefly describe a couple of the threading
		// services provided by the Object class.  All objects have a "monitor"
		// associated with them.  There are a couple of things that a monitor
		// gives you:
		//
		//    1. A synchronization lock.  Only one thread may acquire the lock
		//       at a time.  Any thread that tries to acquire the lock while
		//       another thread has it will block until the lock is released.
		//       You can acquire the lock in two ways: call a synchronized
		//       method on an object (which acquires the lock belonging to the
		//       object you called the method on) or enter a synchronized block
		//       as I have below, which acquires the lock belonging to whatever
		//       object you list in parentheses.
		//
		//    2. A wait/notify mechanism.  A thread may wait for notification
		//       by calling the wait() method on some object.  In that case,
		//       the thread will block until some other thread calls notify()
		//       on the same object.  When notify() is called, exactly one of
		//       the objects that was waiting on the object will wake up and
		//       be allowed to proceed, though you can't be sure which of them
		//       will wake up.  Another scary thing is that, if a thread calls
		//       notify() and no other thread is waiting, the notification is
		//       simply ignored.
		//
		// You don't always need both mechanisms; sometimes locking is all you
		// need.  But there's one subtlety here that bears mentioning.  A thread
		// must have acquired an object's synchronization lock *before* calling
		// wait() or notify().  The penalty is not a compile-time error, but
		// a run-time error (an IllegalMonitorStateException), since a compiler
		// can't reasonably check for this.  (A compiler could check for simple
		// cases, but not more general cases where, for example, the
		// synchronized block is in a method that called the method that is
		// calling notify().)  That's why the call to notify() below is in a
		// synchronized block.  This restriction is a good thing; it helps
		// prevent subtle bugs that arise when one thread calls notify()
		// just before another calls wait().

		synchronized (monitor)
		{
			monitor.notify();
		}


		// Now the main thread will sleep for 15 seconds, allowing all the
		// jabberers to jabber away...

		try
		{
			Thread.sleep(15000);
		}
		catch (InterruptedException e)
		{
		}


		// Then the main thread will interrupt and join all the jabberers,
		// just as we did in the AnnoyingKid example.

		for (Thread t : threads)
		{
			t.interrupt();
		}
		
		try
		{
			for (Thread t : threads)
			{
				t.join();
			}
		}
		catch (InterruptedException e)
		{
			
		}

		System.out.println();
		System.out.println("All the jabberers have stopped.  See ya!");
	}
}
